#!/bin/python
# -*- coding: utf-8 -*-

import os
import sys
import tqdm
import emcee
import numpy as np
from grgrlib import map2arr
from .stats import summary


def run_mcmc(lprob, nsteps, p0=None, moves=None, priors=None, prior_transform=None, backend=None, update_freq=False, resume=False, pool=None, report=None, description=None, temp=1, maintenance_interval=False, verbose=True, **kwargs):
    """Run the emcee sampler.
    """

    if prior_transform is None:
        def prior_transform(x): return x

    if isinstance(backend, str):
        backend = emcee.backends.HDFBackend(
            os.path.splitext(backend)[0] + '.h5')

    if resume:
        p0 = backend.get_chain()[-1]

    nwalks, ndim = np.shape(p0)

    sampler = emcee.EnsembleSampler(
        nwalks, ndim, lprob, moves=moves, pool=pool, backend=backend)

    if not verbose:  # verbose means VERY verbose
        np.warnings.filterwarnings("ignore")

    if verbose > 1:
        report = report or print
    elif verbose:
        pbar = tqdm.tqdm(total=nsteps, unit="sample(s)", dynamic_ncols=True)
        report = report or pbar.write
    else:
        report = lambda x: None

    old_tau = np.inf
    old_lls = np.array([-np.inf]*nwalks)
    cnt = 0

    for result in sampler.sample(p0, iterations=nsteps, **kwargs):

        if verbose == 1:
            lls = list(result)[1]
            maf = 1-sum(lls == old_lls)/nwalks
            try:
                maf = f"{maf:3.0%}"
            except BlockingIOError:
                maf = "??"
            pbar.set_description(
                f"[ll/MAF:{np.max(lls):7.3f}({np.std(lls):1.0e})/{maf}]"
            )
            old_lls = lls.copy()

        if cnt and update_freq and not cnt % update_freq:

            prnttup = f"(mcmc:) Summary from last {update_freq} of {cnt} iterations"

            if temp < 1:
                prnttup += f" with temp of {temp * 100:1.6f}"

            if description is not None:
                prnttup += f" ({str(description)})"

            report(prnttup + ':')

            sample = sampler.get_chain()
            lprobs = sampler.get_log_prob(flat=True)
            acfs = sampler.acceptance_fraction

            tau = emcee.autocorr.integrated_time(sample, tol=0, c=10)
            min_tau = np.min(tau).round(2)
            max_tau = np.max(tau).round(2)

            if priors is not None:
                mcmc_summary(
                    chain=prior_transform(sample[-update_freq:]),
                    lprobs=lprobs[-update_freq:],
                    priors=priors,
                    acceptance_fraction=acfs[-update_freq:],
                    out=lambda x: report(str(x)),
                )

            report(
                f"Autocorrelation times are between {min_tau} and {max_tau}."
            )

        if cnt and update_freq and not (cnt + 1) % update_freq:
            sample = sampler.get_chain()
            old_tau = emcee.autocorr.integrated_time(sample, tol=0)

        if verbose == 1:
            pbar.update(1)

        # avoid mem leakage
        if maintenance_interval and cnt and pool and not cnt % maintenance_interval:
            pool.clear()

        cnt += 1

    if verbose == 1:
        pbar.close()
    if pool:
        pool.close()

    if not verbose:
        np.warnings.filterwarnings("default")

    if backend is None:
        return sampler
    else:
        return backend


def mcmc_summary(
    chain,
    lprobs,
    priors,
    acceptance_fraction=None,
    out=print,
    **args
):

    nchain = chain.reshape(-1, chain.shape[-1])
    lprobs = lprobs.reshape(-1, lprobs.shape[-1])
    mode_x = nchain[lprobs.argmax()]

    res = summary(priors, chain, mode_x, **args)

    out(res.round(3))

    if acceptance_fraction is not None:

        out("Mean acceptance fraction:" +
            str(np.mean(acceptance_fraction).round(3)).rjust(13))

    return res


def get_prior_sample(frozen_prior, nsamples, check_func=False, seed=None, mapper=map, filterwarnings='error', max_attempts=10, debug=False, verbose=True):
    """Get a sample of size `nsamples` from the prior distribution.
    """

    if seed is None:
        seed = np.random.randint(2**31)

    if check_func and not callable(check_func):
        raise Exception('`check_func` must be `False` or a callable')

    def runner(locseed):

        # distribute seed evenly
        np.random.seed((seed + locseed) % 2**31)
        done = False
        no = 0

        while not done:

            no += 1

            with np.warnings.catch_warnings(record=False):
                try:
                    # set warnings
                    np.warnings.filterwarnings(filterwarnings)
                    rst = np.random.randint(2 ** 31)  # win explodes with 2**32
                    # draw from prior
                    pdraw = [
                        pl.rvs(random_state=rst + sn)
                        for sn, pl in enumerate(frozen_prior)
                    ]

                    # check if function returns finite results (if provided)
                    if check_func:
                        draw_prob = check_func(pdraw)
                        done = not np.any(np.isinf(draw_prob))
                    else:
                        done = True

                # be kind to errors
                except Exception as e:
                    if debug:
                        print(str(e) + f" ({no})")
                    if not locseed and no == max_attempts:
                        raise type(e)(
                            str(e) + f" (after {no} unsuccessful attemps).").with_traceback(sys.exc_info()[2])
                    else:
                        pass

        return pdraw, no

    if verbose > 1:
        print("(prior_sample:) Sampling from the pior...")

    wrapper = tqdm.tqdm if verbose < 2 else (lambda x, **kwarg: x)
    pmap_sim = wrapper(mapper(runner, range(nsamples)), total=nsamples)

    draws, nos = map2arr(pmap_sim)

    if verbose and check_func:
        print(
            f"(prior_sample:) Sampling done. Check fails for {100 * (sum(nos) - nsamples) / sum(nos):2.2f}% of the prior.")

    return draws



#!/bin/python
# -*- coding: utf-8 -*-

import os
import numpy as np
from .moves import DIMEMove
from .sampler import run_mcmc
from scipy.stats import multivariate_normal, norm
from scipy.special import logsumexp
from emcee.tests.integration.test_proposal import _test_normal, _test_uniform

filepath = os.path.dirname(__file__)


def _create_test_func(ndim, weight, distance, cov_scale):

    cov = np.eye(ndim)*cov_scale
    mean = np.zeros(ndim)
    mean[0] = distance

    lw0 = np.log(weight[0])
    lw1 = np.log(weight[1])
    lw2 = np.log(1-weight[0]-weight[1])

    dist = multivariate_normal(np.zeros(ndim), cov)

    def log_prob(p):
        return logsumexp((lw0 + dist.logpdf(p + mean), lw1 + dist.logpdf(p), lw2 + dist.logpdf(p - mean)))

    return log_prob


def _marginal_pdf_test_func(x, cov_scale, m, weight):

    normal = norm(scale=np.sqrt(cov_scale))

    return weight[0]*normal.pdf(x+m) + weight[1]*normal.pdf(x) + (1-weight[0]-weight[1])*normal.pdf(x-m)


def test_example(create=False, seed=1234):

    np.random.seed(seed)

    # define distribution
    m = 2
    cov_scale = 0.05
    weight = (0.33, 0.1)
    ndim = 35

    log_prob = _create_test_func(ndim, weight, m, cov_scale)

    nchain = ndim*5
    niter = 300

    initmean = np.zeros(ndim)
    initcov = np.eye(ndim)*np.sqrt(2)
    initchain = multivariate_normal(
        mean=initmean, cov=initcov, seed=seed).rvs(nchain)

    moves = DIMEMove(aimh_prob=.1, df_proposal_dist=10)
    sampler = run_mcmc(log_prob, niter, p0=initchain,
                       moves=moves, verbose=create)
    chain = sampler.get_chain()

    path = os.path.join(filepath, "test_storage", "median.npy")
    median = np.median(chain[-int(niter/3):, :, 0])

    if create:
        np.save(path, median)
        print(f'Test file updated at {path} with {median}')
    else:
        test_median = np.load(path)
        np.testing.assert_allclose(median, test_median, atol=1e-4)


def test_normal_dime(**kwargs):
    _test_normal(DIMEMove(), **kwargs)


def test_uniform_dime(**kwargs):
    _test_uniform(DIMEMove(), **kwargs)



# -*- coding: utf-8 -*-

from emcee.moves.red_blue import RedBlueMove
from emcee.moves.de import DEMove
from scipy.special import logsumexp
import numpy as np
import scipy.stats as ss


def mvt_sample(df, mean, cov, size, random):
    """Sample from multivariate t distribution

    For reasons beyond my understanding, the results from random.multivariate_normal with non-identity covariance matrix are not reproducibel across architecture. Since scipy.stats.multivariate_t is based on numpy's multivariate_normal, the workaround is to crochet this manually. Advantage is that the scipy dependency drops out.
    """

    dim = len(mean)

    # draw samples
    snorm = random.randn(size, dim)
    chi2 = random.chisquare(df, size) / df

    # calculate sqrt of covariance
    svd_cov = np.linalg.svd(cov * (df - 2) / df)
    sqrt_cov = svd_cov[0] * np.sqrt(svd_cov[1]) @ svd_cov[2]

    return mean + snorm @ sqrt_cov / np.sqrt(chi2)[:, None]


class DIMEMove(RedBlueMove):
    r"""A proposal using adaptive differential-independence mixture ensemble MCMC.

    This is the `Differential-Independence Mixture Ensemble proposal` as developed in `Ensemble MCMC Sampling for DSGE Models <https://gregorboehl.com/live/ademc_boehl.pdf>`_ (previousy ADEMC).

    Parameters
    ----------
    sigma : float, optional
        standard deviation of the Gaussian used to stretch the proposal vector.
    gamma : float, optional
        mean stretch factor for the proposal vector. By default, it is :math:`2.38 / \sqrt{2\,\mathrm{ndim}}` as recommended by `ter Braak (2006) <http://www.stat.columbia.edu/~gelman/stuff_for_blog/cajo.pdf>`_.
    aimh_prob : float, optional
        probability to draw an adaptive independence Metropolis Hastings (AIMH) proposal. By default this is set to :math:`0.1`.
    df_proposal_dist : float
        degrees of freedom of the multivariate t distribution used for AIMH proposals. Defaults to :math:`10`.
    """

    def __init__(
        self, sigma=1.0e-5, gamma=None, aimh_prob=0.1, df_proposal_dist=10, **kwargs
    ):

        self.sigma = sigma
        self.g0 = gamma
        self.aimh_prob = aimh_prob
        self.dft = df_proposal_dist

        kwargs["nsplits"] = 1
        super(DIMEMove, self).__init__(**kwargs)

    def setup(self, coords):
        # set some sane defaults

        nchain, npar = coords.shape

        if self.g0 is None:
            # pure MAGIC
            self.g0 = 2.38 / np.sqrt(2 * npar)

        if not hasattr(self, "prop_cov"):
            # even more MAGIC
            self.prop_cov = np.eye(npar)
            self.prop_mean = np.zeros(npar)
            self.accepted = np.ones(nchain, dtype=bool)
            self.cumlweight = -np.inf

    def propose(self, model, state):
        # wrap original propose to get the some info on the current state
        self.lprobs = state.log_prob
        state, accepted = super(DIMEMove, self).propose(model, state)
        self.accepted = accepted
        return state, accepted

    def get_proposal(self, x, dummy, random):
        # actual proposal function

        nchain, npar = x.shape

        # differential evolution: draw the indices of the complementary chains
        i0 = np.arange(nchain) + random.randint(1, nchain, size=nchain)
        i1 = np.arange(nchain) + random.randint(1, nchain - 1, size=nchain)
        i1[i1 >= i0] += 1
        # add small noise and calculate proposal
        f = self.sigma * random.randn(nchain)
        q = x + self.g0 * (x[i0 % nchain] - x[i1 % nchain]) + f[:, np.newaxis]
        factors = np.zeros(nchain, dtype=np.float64)

        # log weight of current ensemble
        lweight = logsumexp(self.lprobs) + \
            np.log(sum(self.accepted)) - np.log(nchain)

        # calculate stats for current ensemble
        ncov = np.cov(x.T, ddof=1)
        nmean = np.mean(x, axis=0)

        # update AIMH proposal distribution
        newcumlweight = np.logaddexp(self.cumlweight, lweight)
        self.prop_cov = (
            np.exp(self.cumlweight - newcumlweight) * self.prop_cov
            + np.exp(lweight - newcumlweight) * ncov
        )
        self.prop_mean = (
            np.exp(self.cumlweight - newcumlweight) * self.prop_mean
            + np.exp(lweight - newcumlweight) * nmean
        )
        self.cumlweight = newcumlweight

        # draw chains for AIMH sampling
        xchnge = random.rand(nchain) <= self.aimh_prob

        # draw alternative candidates and calculate their proposal density
        xcand = mvt_sample(
            df=self.dft,
            mean=self.prop_mean,
            cov=self.prop_cov,
            size=sum(xchnge),
            random=random,
        )
        lprop_old = ss.multivariate_t.logpdf(
            x[xchnge],
            self.prop_mean,
            self.prop_cov * (self.dft - 2) / self.dft,
            df=self.dft,
        )
        lprop_new = ss.multivariate_t.logpdf(
            xcand,
            self.prop_mean,
            self.prop_cov * (self.dft - 2) / self.dft,
            df=self.dft,
        )

        # update proposals and factors
        q[xchnge, :] = np.reshape(xcand, (-1, npar))
        factors[xchnge] = lprop_old - lprop_new

        return q, factors


ADEMove = DIMEMove  # set alias for backwards compatibility



#!/bin/python
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib.lines import Line2D
from .stats import calc_min_interval as hpd


def fast_kde(x, bw=4.5):
    """
    A fft-based Gaussian kernel density estimate (KDE)
    The code was adapted from https://github.com/mfouesneau/faststats

    Parameters
    ----------
    x : Numpy array or list
    bw : float
        Bandwidth scaling factor for the KDE. Should be larger than 0. The higher this number the
        smoother the KDE will be. Defaults to 4.5 which is essentially the same as the Scott's rule
        of thumb (the default rule used by SciPy).

    Returns
    -------
    density: A gridded 1D KDE of the input points (x)
    xmin: minimum value of x
    xmax: maximum value of x
    """

    from scipy.signal import gaussian, convolve
    from scipy.stats import entropy

    x = np.asarray(x, dtype=float)
    x = x[np.isfinite(x)]
    n = len(x)
    nx = 200

    xmin, xmax = np.min(x), np.max(x)

    dx = (xmax - xmin) / (nx - 1)
    std_x = entropy((x - xmin) / dx) * bw
    if ~np.isfinite(std_x):
        std_x = 0.0
    grid, _ = np.histogram(x, bins=nx)

    scotts_factor = n ** (-0.2)
    kern_nx = int(scotts_factor * 2 * np.pi * std_x)
    kernel = gaussian(kern_nx, scotts_factor * std_x)

    npad = min(nx, 2 * kern_nx)
    grid = np.concatenate([grid[npad:0:-1], grid, grid[nx: nx - npad: -1]])
    density = convolve(grid, kernel, mode="same")[npad: npad + nx]

    norm_factor = n * dx * (2 * np.pi * std_x ** 2 * scotts_factor ** 2) ** 0.5

    density = density / norm_factor

    return density, xmin, xmax


def kdeplot_op(ax, data, bw, prior=None, prior_alpha=1, prior_style="--"):
    """Get a list of density and likelihood plots, if a prior is provided."""
    ls = []
    pls = []
    errored = []
    density, l, u = fast_kde(data, bw)
    x = np.linspace(l, u, len(density))
    if prior is not None:

        try:
            p = prior.logpdf(x)
        except ValueError:
            p = x.copy()
            for i, xi in enumerate(x):
                p[i] = prior.logpdf(xi)
        pls.append(ax.plot(x, np.exp(p), alpha=prior_alpha, ls=prior_style))

    ls.append(ax.plot(x, density))

    if errored:
        ax.text(
            0.27,
            0.47,
            "WARNING: KDE plot failed for: " + ",".join(errored),
            bbox={"facecolor": "red", "alpha": 0.5, "pad": 10},
            style="italic",
        )

    return ls, pls


def get_axis(ax, default_rows, default_columns, **default_kwargs):
    """Verifies the provided axis is of the correct shape, and creates one if needed.

    Args:
        ax: matplotlib axis or None
        default_rows: int, expected rows in axis
        default_columns: int, expected columns in axis
        **default_kwargs: keyword arguments to pass to plt.subplot

    Returns:
        axis, or raises an error
    """

    default_shape = (default_rows, default_columns)
    if ax is None:
        fig, ax = plt.subplots(*default_shape, **default_kwargs)
    elif ax.shape != default_shape:
        raise ValueError("Subplots with shape %r required" % (default_shape,))
    return fig, ax


def plot_posterior_op(
    trace_values,
    ax,
    bw,
    prior,
    kde_plot=False,
    point_estimate="mean",
    round_to=3,
    alpha_level=0.05,
    ref_val=None,
    rope=None,
    text_size=16,
    display_additinal_info=False,
    **kwargs
):
    """Artist to draw posterior."""

    def format_as_percent(x, round_to=0):
        return "{0:.{1:d}f}%".format(100 * x, round_to)

    def display_ref_val(ref_val):
        less_than_ref_probability = (trace_values < ref_val).mean()
        greater_than_ref_probability = (trace_values >= ref_val).mean()
        ref_in_posterior = "{} <{:g}< {}".format(
            format_as_percent(less_than_ref_probability, 1),
            ref_val,
            format_as_percent(greater_than_ref_probability, 1),
        )
        ax.axvline(ref_val, bottom=0.02, ymax=0.75,
                   color="g", linewidth=4, alpha=0.65)
        ax.text(
            trace_values.mean(),
            plot_height * 0.6,
            ref_in_posterior,
            size=text_size,
            horizontalalignment="center",
        )

    def display_rope(rope):
        ax.plot(
            rope,
            (plot_height * 0.02, plot_height * 0.02),
            linewidth=20,
            color="r",
            alpha=0.75,
        )
        text_props = dict(
            size=text_size, horizontalalignment="center", color="r")
        ax.text(rope[0], plot_height * 0.14, rope[0], **text_props)
        ax.text(rope[1], plot_height * 0.14, rope[1], **text_props)

    def display_point_estimate():
        if not point_estimate:
            return
        if point_estimate not in ("mode", "mean", "median"):
            raise ValueError(
                "Point Estimate should be in ('mode','mean','median')")
        if point_estimate == "mean":
            point_value = trace_values.mean()
        elif point_estimate == "mode":
            if isinstance(trace_values[0], float):
                density, l, u = fast_kde(trace_values, bw)
                x = np.linspace(l, u, len(density))
                point_value = x[np.argmax(density)]
            else:
                point_value = mode(trace_values.round(round_to))[0][0]
        elif point_estimate == "median":
            point_value = np.median(trace_values)
        point_text = "{point_estimate}={point_value:.{round_to}f}".format(
            point_estimate=point_estimate, point_value=point_value, round_to=round_to
        )

        ax.text(
            point_value,
            plot_height * 0.8,
            point_text,
            size=text_size,
            horizontalalignment="center",
        )

    def display_hpd():
        sorted_trace_values = np.sort(trace_values)
        hpd_intervals = hpd(sorted_trace_values, alpha=alpha_level)
        ax.plot(
            hpd_intervals,
            (plot_height * 0.02, plot_height * 0.02),
            linewidth=4,
            color="k",
        )
        ax.text(
            hpd_intervals[0],
            plot_height * 0.07,
            hpd_intervals[0].round(round_to),
            size=text_size,
            horizontalalignment="right",
        )
        ax.text(
            hpd_intervals[1],
            plot_height * 0.07,
            hpd_intervals[1].round(round_to),
            size=text_size,
            horizontalalignment="left",
        )
        ax.text(
            (hpd_intervals[0] + hpd_intervals[1]) / 2,
            plot_height * 0.2,
            format_as_percent(1 - alpha_level) + " HPD",
            size=text_size,
            horizontalalignment="center",
        )

    def format_axes():
        ax.yaxis.set_ticklabels([])
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.spines["bottom"].set_visible(True)
        ax.yaxis.set_ticks_position("none")
        ax.xaxis.set_ticks_position("bottom")
        # ax.tick_params(axis='x', direction='out', width=1, length=3,
        # color='0.5', labelsize=text_size)
        ax.spines["bottom"].set_color("0.5")

    def set_key_if_doesnt_exist(d, key, value):
        if key not in d:
            d[key] = value

    if kde_plot and isinstance(trace_values[0], float):
        kdeplot_op(
            ax, trace_values, bw=bw, prior_alpha=kwargs.pop("alpha", 0.35), **kwargs
        )

    else:
        set_key_if_doesnt_exist(kwargs, "bins", 30)
        set_key_if_doesnt_exist(kwargs, "edgecolor", "w")
        set_key_if_doesnt_exist(kwargs, "align", "right")
        set_key_if_doesnt_exist(kwargs, "color", "#87ceeb")
        ax.hist(trace_values, density=True, **kwargs)

    if prior is not None:
        x = np.linspace(*ax.get_xlim(), 100)
        ax.plot(x, prior.pdf(x), "--", c="C1", alpha=0.8)

    plot_height = ax.get_ylim()[1]

    format_axes()

    if display_additinal_info:
        display_hpd()
        display_point_estimate()
        if ref_val is not None:
            display_ref_val(ref_val)
        if rope is not None:
            display_rope(rope)


def scale_text(figsize, text_size):
    """Scale text to figsize."""

    if text_size is None and figsize is not None:
        if figsize[0] <= 11:
            return 12
        else:
            return figsize[0]
    else:
        return text_size


def posteriorplot(
    trace,
    varnames=None,
    tune=0,
    figsize=None,
    plots_per_fig=4,
    text_size=None,
    ropep=None,
    ref_val=None,
    round_to=3,
    alpha_level=0.05,
    bw=4.5,
    ax=None,
    **kwargs
):

    axs = []
    figs = []

    if varnames is not None:
        dim = len(varnames)
    else:
        dim = trace.shape[-1]

    for ic in range(0, dim, plots_per_fig):

        if varnames is not None:
            vnames_chunk = varnames[ic: ic + plots_per_fig]
        else:
            vnames_chunk = [None] * min(plots_per_fig, dim - ic)
        trace_chunk = trace[..., ic: ic + plots_per_fig]

        def create_axes_grid(figsize, traces):
            l_trace = len(traces)
            if l_trace == 1:
                fig, ax = plt.subplots(figsize=figsize)
            else:
                n = np.ceil(l_trace / 2.0).astype(int)
                if figsize is None:
                    figsize_loc = (8, len(vnames_chunk) * 1.5)
                fig, ax = plt.subplots(n, 2, figsize=figsize_loc)
                ax = ax.reshape(2 * n)
                if l_trace % 2 == 1:
                    ax[-1].set_axis_off()
                    ax = ax[:-1]
            return fig, ax

        if ax is None:
            fig, ax = create_axes_grid(figsize, vnames_chunk)
        else:
            ax = apx

        var_num = len(vnames_chunk)
        if ref_val is None:
            ref_val = [None] * var_num
        elif np.isscalar(ref_val):
            ref_val = [ref_val for _ in range(var_num)]

        if ropep is None:
            rope = [None] * var_num
        elif np.ndim(ropep) == 1:
            rope = [rope] * var_num

        if len(np.atleast_1d(ax).flatten()) != len(vnames_chunk):
            print("Given axis does not match number of plots")
        for idx, (a, v) in enumerate(zip(np.atleast_1d(ax), vnames_chunk)):
            tr_values = trace_chunk[-tune:, :, idx].flatten()
            plot_posterior_op(
                tr_values,
                ax=a,
                bw=bw,
                prior=None,
                round_to=round_to,
                alpha_level=alpha_level,
                ref_val=ref_val[idx],
                rope=rope[idx],
                text_size=scale_text(figsize, text_size),
                **kwargs
            )
            a.set_title(v, fontsize=scale_text(figsize, text_size))

        plt.tight_layout()

        axs.append(ax)
        figs.append(fig)

    return figs, axs


def traceplot(
    trace,
    varnames,
    tune,
    figsize=None,
    plots_per_fig=3,
    frozen_prior=None,
    draw_each_trace=True,
    bw=4.5,
    text_size=None,
    **kwargs
):

    # inspired by pymc3 with kisses

    if figsize is None:
        figsize = 9, plots_per_fig * 2.5

    width = trace.shape[0]

    if frozen_prior is None:
        frozen_prior = [None] * len(varnames)

    custom_lines_hist = [
        Line2D([0], [0], linestyle="-", color="C0", lw=1),
        Line2D([0], [0], linestyle="--", color="C1", lw=1),
    ]

    custom_lines_trace = [
        Line2D([0], [0], linestyle="-", color="maroon", lw=1),
        Line2D([0], [0], linestyle="-", color="C0", lw=1),
    ]

    axs = []
    figs = []
    subfigs = []

    for chunk in range(0, len(varnames), plots_per_fig):

        eff_ppf = min(plots_per_fig, len(varnames[chunk:]))
        eff_figsize = figsize[0], figsize[1] * eff_ppf / plots_per_fig

        figs.append(plt.figure(constrained_layout=True, figsize=eff_figsize))
        subfig = figs[-1].subfigures(eff_ppf, 1, wspace=0.07)

        if eff_ppf == 1:
            subfig = [subfig]

        subfigs.append(subfig)

        for i in range(eff_ppf):

            axs.append(subfigs[-1][i].subplots(1, 2))
            data = trace[..., chunk + i]

            posterior = data[-tune:].flatten()
            plot_posterior_op(
                posterior,
                ax=axs[-1][0],
                bw=bw,
                prior=frozen_prior[chunk + i],
                text_size=scale_text(figsize, text_size),
                **kwargs
            )
            if draw_each_trace:
                axs[-1][1].plot(
                    range(0, width - tune + 1), data[: - tune + 1], c="maroon", alpha=0.03
                )
                axs[-1][1].plot(range(width - tune, width),
                                data[- tune:], c="C0", alpha=0.045)
                axs[-1][1].plot(
                    [width - tune, width - tune],
                    [
                        np.mean(data, 1)[- tune] - np.std(data, 1)[- tune] * 3,
                        np.mean(data, 1)[- tune] + np.std(data, 1)[- tune] * 3,
                    ],
                    "--",
                    alpha=0.4,
                    color="k",
                )

            else:
                i95s = np.percentile(data, [2.5, 97.5], axis=1)
                i66s = np.percentile(data, [17, 83], axis=1)
                means = np.mean(data, axis=1)
                medis = np.median(data, axis=1)

                axs[-1][1].fill_between(
                    range(0, width - tune + 1),
                    *i95s[:, : - tune + 1],
                    lw=0,
                    alpha=0.1,
                    color="C1"
                )
                axs[-1][1].fill_between(
                    range(width - tune, width), *i95s[:, - tune:], lw=0, alpha=0.2, color="C1"
                )
                axs[-1][1].fill_between(
                    range(0, width - tune + 1),
                    *i66s[:, : - tune + 1],
                    lw=0,
                    alpha=0.3,
                    color="C1"
                )
                axs[-1][1].fill_between(
                    range(width - tune, width), *i66s[:, - tune:], lw=0, alpha=0.4, color="C1"
                )
                axs[-1][1].plot(range(width - tune, width), means[- tune:], lw=2, c="C0")
                axs[-1][1].plot(
                    range(0, width - tune + 1), means[: - tune + 1], lw=2, c="C0", alpha=0.5
                )

                axs[-1][1].plot(
                    [width - tune, width - tune],
                    [
                        np.mean(data, 1)[- tune] - np.std(data, 1)[- tune] * 3,
                        np.mean(data, 1)[- tune] + np.std(data, 1)[- tune] * 3,
                    ],
                    "--",
                    alpha=0.4,
                    color="k",
                )

            axs[-1][0].tick_params(
                axis="x",
                direction="out",
                width=1,
                length=3,
                color="0.5",
                labelsize=scale_text(figsize, text_size),
            )
            axs[-1][1].tick_params(
                axis="x",
                direction="out",
                width=1,
                length=3,
                color="0.5",
                labelsize=scale_text(figsize, text_size),
            )
            axs[-1][1].tick_params(
                axis="y",
                width=1,
                length=0,
                color="0.5",
                labelsize=scale_text(figsize, text_size),
            )

            axs[-1][0].set_ylabel("Frequency")
            axs[-1][1].set_ylabel("Sample value")
            axs[-1][0].set_ylim(bottom=0)

            subfigs[-1][i].suptitle(str(varnames[chunk + i]))

        if frozen_prior[chunk] is not None:
            axs[-1][0].legend(custom_lines_hist, ["Posterior", "Prior"])
            axs[-1][1].legend(custom_lines_trace, ["Burn-in", "Posterior"])

    return figs, subfigs, axs





#!/bin/python
# -*- coding: utf-8 -*-

import numpy as np
import scipy.stats as ss
from scipy.special import gammaln


class InvGammaDynare(ss.rv_continuous):

    name = "inv_gamma_dynare"

    # @np.vectorize
    def _logpdf(self, x, s, nu):

        x = np.atleast_1d(x)
        s = np.atleast_1d(s)
        nu = np.atleast_1d(nu)

        lpdf = np.empty_like(x)
        lpdf[:] = -np.inf
        lpdf[x >= 0] = (
            np.log(2)
            - gammaln(nu[x >= 0] / 2)
            - nu[x >= 0] / 2 * (np.log(2) - np.log(s[x >= 0]))
            - (nu[x >= 0] + 1) * np.log(x[x >= 0])
            - 0.5 * s[x >= 0] / np.square(x[x >= 0])
        )

        return lpdf

    def _pdf(self, x, s, nu):
        return np.exp(self._logpdf(x, s, nu))


def inv_gamma_spec(mu, sigma):

    # directly taken from dynare/matlab

    def ig1fun(nu):
        return (
            np.log(2 * mu ** 2)
            - np.log((sigma ** 2 + mu ** 2) * (nu - 2))
            + 2 * (gammaln(nu / 2) - gammaln((nu - 1) / 2))
        )

    nu = np.sqrt(2 * (2 + mu ** 2 / sigma ** 2))
    nu2 = 2 * nu
    nu1 = 2
    err = ig1fun(nu)
    err2 = ig1fun(nu2)

    if err2 > 0:
        while nu2 < 1e12:  # Shift the interval containing the root.
            nu1 = nu2
            nu2 = nu2 * 2
            err2 = ig1fun(nu2)
            if err2 < 0:
                break
        if err2 > 0:
            raise ValueError(
                "[inv_gamma_spec:] Failed in finding an interval containing a sign change! You should check that the prior variance is not too small compared to the prior mean..."
            )

    # Solve for nu using the secant method.
    while abs(nu2 / nu1 - 1) > 1e-14:
        if err > 0:
            nu1 = nu
            if nu < nu2:
                nu = nu2
            else:
                nu = 2 * nu
                nu2 = nu
        else:
            nu2 = nu
        nu = (nu1 + nu2) / 2
        err = ig1fun(nu)

    s = (sigma ** 2 + mu ** 2) * (nu - 2)

    if (
        abs(
            np.log(mu)
            - np.log(np.sqrt(s / 2))
            - gammaln((nu - 1) / 2)
            + gammaln(nu / 2)
        )
        > 1e-7
    ):
        raise ValueError(
            "[inv_gamma_spec:] Failed in solving for the hyperparameters!")
    if abs(sigma - np.sqrt(s / (nu - 2) - mu * mu)) > 1e-7:
        raise ValueError(
            "[inv_gamma_spec:] Failed in solving for the hyperparameters!")

    return s, nu



#!/bin/python
# -*- coding: utf-8 -*-

import warnings
import emcee
import numpy as np
import pandas as pd
import scipy.stats as ss

try:
    from grgrlib.njitted import mvn_logpdf as logpdf
except ModuleNotFoundError:
    def logpdf(drv, cmean, ccov): return ss.multivariate_normal(
        mean=cmean, cov=ccov).logpdf(drv)


def mdd_laplace(chain, lprobs, calc_hess=False):
    """Approximate the marginal data density useing the LaPlace method."""

    if chain.ndim > 2:
        chain = chain.reshape(-1, chain.shape[2])

    lprobs = lprobs.flatten()

    mode_x = chain[lprobs.argmax()]

    if calc_hess:

        try:
            from numdifftools import Hessian
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "`mdd_laplace` with `calc_hess=True` requires the numdifftools package to be installed.")

        np.warnings.filterwarnings("ignore")
        hh = Hessian(func)(mode_x)
        np.warnings.filterwarnings("default")

        if np.isnan(hh).any():
            raise ValueError(
                "[mdd:]".ljust(15, " ")
                + "Option `hess` is experimental and did not return a usable hessian matrix."
            )

        inv_hess = np.linalg.inv(hh)

    else:
        inv_hess = np.cov(chain.T)

    ndim = chain.shape[-1]
    log_det_inv_hess = np.log(np.linalg.det(inv_hess))
    mdd = 0.5 * ndim * np.log(2 * np.pi) + 0.5 * \
        log_det_inv_hess + lprobs.max()

    return mdd


def mdd_harmonic_mean(chain, lprobs, pool=None, alpha=0.05, verbose=False, debug=False):
    """Approximate the marginal data density useing modified harmonic mean."""

    if chain.ndim > 2:
        chain = chain.reshape(-1, chain.shape[2])

    lprobs = lprobs.flatten()

    cmean = chain.mean(axis=0)
    ccov = np.cov(chain.T)
    cicov = np.linalg.inv(ccov)

    nsamples = chain.shape[0]

    def runner(chunk):

        res = np.empty_like(chunk)
        wrapper = tqdm.tqdm if verbose else (lambda x, **kwarg: x)

        for i in wrapper(range(len(chunk))):

            drv = chain[i]
            drl = lprobs[i]

            if (drv - cmean) @ cicov @ (drv - cmean) < ss.chi2.ppf(
                1 - alpha, df=chain.shape[-1]
            ):
                res[i] = logpdf(drv, cmean, ccov) - drl
            else:
                res[i] = -np.inf

        return res

    if not debug and pool is not None:
        nbatches = pool.ncpus
        batches = pool.imap(runner, np.array_split(chain, nbatches))
        mls = np.vstack(list(batches))
    else:
        mls = runner(chain)

    maxllike = np.max(mls)  # for numeric stability
    imdd = np.log(np.mean(np.exp(mls - maxllike))) + maxllike

    return -imdd


def calc_min_interval(x, alpha):
    """Internal method to determine the minimum interval of
    a given width

    Assumes that x is sorted numpy array.
    """
    n = len(x)
    cred_mass = 1.0 - alpha

    interval_idx_inc = int(np.floor(cred_mass * n))
    n_intervals = n - interval_idx_inc
    interval_width = x[interval_idx_inc:] - x[:n_intervals]

    if len(interval_width) == 0:
        # raise ValueError('Too few elements for interval calculation')
        warnings.warn("Too few elements for interval calculation.")

        return None, None

    else:
        min_idx = np.argmin(interval_width)
        hdi_min = x[min_idx]
        hdi_max = x[min_idx + interval_idx_inc]

        return hdi_min, hdi_max


def mc_error(x):
    means = np.mean(x, 0)
    return np.std(means) / np.sqrt(x.shape[0])


def _hpd_df(x, alpha):

    cnames = [
        "hpd_{0:g}".format(100 * alpha / 2),
        "hpd_{0:g}".format(100 * (1 - alpha / 2)),
    ]

    sx = np.sort(x.flatten())
    hpd_vals = np.array(calc_min_interval(sx, alpha)).reshape(1, -1)

    return pd.DataFrame(hpd_vals, columns=cnames)


def gelman_rubin(sample):
    """The Gelman-Rubin (1992) statistics.
    """

    ndraws, nchains, npars = sample.shape

    chain_mean = sample.mean(axis=0)
    between_chain_var = ndraws*np.var(chain_mean, axis=0, ddof=1)
    within_chain_var = np.var(sample, axis=0, ddof=1)
    mean_variance = 1/nchains*np.sum(within_chain_var, axis=0)

    return 1 - 1/ndraws + 1/ndraws*between_chain_var/mean_variance


def summary(priors, store, pmode=None, bounds=None, alpha=0.1, top=None, show_prior=True):
    # inspired by pymc3 because it looks really nice

    if bounds is not None or isinstance(store, tuple):
        xs, fs, ns = store
        ns = ns.squeeze()
        fas = (-fs[:, 0]).argsort()
        xs = xs[fas]
        fs = fs.squeeze()[fas]

    f_prs = [
        lambda x: pd.Series(x, name="distribution"),
        lambda x: pd.Series(x, name="pst_mean"),
        lambda x: pd.Series(x, name="sd/df"),
    ]

    f_bnd = [
        lambda x: pd.Series(x, name="lbound"),
        lambda x: pd.Series(x, name="ubound"),
    ]

    def mode_func(x, n):
        return pmode[n] if pmode is not None else mode(x.flatten())

    funcs = [
        lambda x, n: pd.Series(np.mean(x), name="mean"),
        lambda x, n: pd.Series(np.std(x), name="sd"),
        lambda x, n: pd.Series(
            mode_func(x, n), name="mode" if pmode is not None else "marg. mode"
        ),
        lambda x, n: _hpd_df(x, alpha),
        lambda x, n: pd.Series(mc_error(x), name="error"),
    ]

    var_dfs = []
    for i, var in enumerate(priors):

        lst = []
        if show_prior:
            prior = priors[var]
            if len(prior) > 3:
                prior = prior[-3:]
            [lst.append(f(prior[j])) for j, f in enumerate(f_prs)]
            if bounds is not None:
                [lst.append(f(np.array(bounds).T[i][j]))
                 for j, f in enumerate(f_bnd)]

        if bounds is not None:
            [lst.append(pd.Series(s[i], name=n))
             for s, n in zip(xs[:top], ns[:top])]
        else:
            vals = store[:, :, i]
            [lst.append(f(vals, i)) for f in funcs]
        var_df = pd.concat(lst, axis=1)
        var_df.index = [var]
        var_dfs.append(var_df)

    if bounds is not None:

        lst = []

        if show_prior:
            [lst.append(f("")) for j, f in enumerate(f_prs)]
            if bounds is not None:
                [lst.append(f("")) for j, f in enumerate(f_bnd)]

        [lst.append(pd.Series(s, name=n)) for s, n in zip(fs[:top], ns[:top])]
        var_df = pd.concat(lst, axis=1)
        var_df.index = ["loglike"]
        var_dfs.append(var_df)

    dforg = pd.concat(var_dfs, axis=0, sort=False)

    return dforg


# a wrapper without warning
def integrated_time(x): return emcee.autocorr.integrated_time(x, tol=0)



#!/bin/python
# -*- coding: utf-8 -*-

import os
import emcee
import numpy as np
import scipy.stats as ss
import scipy.optimize as so
from grgrlib import parse_yaml
from scipy.special import logit, expit
from .dists import inv_gamma_spec, InvGammaDynare


def get_prior(prior, verbose=False):
    """Compile prior-related computational objects from a list of priors.
    """

    prior_lst = ()
    initv, lb, ub = [], [], []
    funcs_con, funcs_re = (), () # prior-to-sampler, sampler-to-prior

    snorm = ss.norm()

    if verbose:
        print("Adding parameters to the prior distribution...")

    for pp in prior:

        dist = prior[str(pp)]

        if len(dist) == 3:
            initv.append(None)
            lb.append(None)
            ub.append(None)
            ptype, pmean, pstdd = dist
        elif len(dist) == 6:
            initv.append(eval(str(dist[0])))
            lb.append(dist[1])
            ub.append(dist[2])
            ptype, pmean, pstdd = dist[3:]
        else:
            raise NotImplementedError(
                "Prior specification must either be 3 or 6 inputs but is %s." % pp)

        # simply make use of frozen distributions
        if str(ptype) == "uniform":
            ndist = ss.uniform(loc=pmean, scale=pstdd - pmean)

        elif str(ptype) == "normal":
            ndist = ss.norm(loc=pmean, scale=pstdd)

        elif str(ptype) == "gamma":
            b = pstdd ** 2 / pmean
            a = pmean / b
            ndist = ss.gamma(a, scale=b)

        elif str(ptype) == "beta":
            a = (1 - pmean) * pmean ** 2 / pstdd ** 2 - pmean
            b = a * (1 / pmean - 1)
            ndist = ss.beta(a=a, b=b)

        elif str(ptype) == "inv_gamma":

            def targf(x):
                y0 = ss.invgamma(x[0], scale=x[1]).std() - pstdd
                y1 = ss.invgamma(x[0], scale=x[1]).mean() - pmean
                return np.array([y0, y1])

            ig_res = so.root(targf, np.array([4, 4]), method="lm")

            if ig_res["success"] and np.allclose(targf(ig_res["x"]), 0):
                ndist = ss.invgamma(ig_res["x"][0], scale=ig_res["x"][1])
            else:
                raise ValueError(
                    f"Can not find inverse gamma distribution with mean {pmean} and std {pstdd}.")

        elif str(ptype) == "inv_gamma_dynare":
            s, nu = inv_gamma_spec(pmean, pstdd)
            ndist = InvGammaDynare()(s, nu)

        else:
            raise NotImplementedError(
                f" Distribution {ptype} not implemented.")

        prior_lst += ndist,
        if str(ptype) in ('gamma', 'inv_gamma', 'inv_gamma_dynare'):
            funcs_re += np.log,
            funcs_con += np.exp,
        elif str(ptype) == 'beta':
            funcs_re += logit,
            funcs_con += expit,
        else:
            funcs_re += lambda x: x,
            funcs_con += lambda x: x,

        if verbose:
            if len(dist) == 3:
                print(
                    "   - %s as %s with mean %s and std/df %s"
                    % (pp, ptype, pmean, pstdd)
                )
            if len(dist) == 6:
                print(
                    "   - %s as %s (%s, %s). Init @ %s, with bounds (%s, %s)"
                    % (pp, ptype, pmean, pstdd, dist[0], dist[1], dist[2])
                )

    return list(prior_lst), get_log_prior(prior_lst), get_bijective_prior_transformation(funcs_con, funcs_re), initv, (lb, ub)


def get_log_prior(frozen_prior):
    """Get the log-prior function.
    """

    def log_prior(par):

        prior = 0
        for i, pl in enumerate(frozen_prior):
            prior += pl.logpdf(par[i])

        return prior

    return log_prior


def get_bijective_prior_transformation(funcs_con, funcs_re):
    """Get the bijective prior transformation function.
    """

    def bijective_prior_transformation(x, sampler_to_prior=True):

        x = np.array(x)
        res = x.copy()

        for i in range(x.shape[-1]):
            res[...,i] = funcs_con[i](x[...,i]) if sampler_to_prior else funcs_re[i](x[...,i])

        return res

    return bijective_prior_transformation


def find_mode_simple(lprob, init, frozen_prior, sd=True, verbose=False, **kwargs):

    def objective(x):

        ll = -lprob(x) - log_prior(x, frozen_prior)

        if verbose:
            print(-ll)

        return ll

    if not 'method' in kwargs:
        kwargs['method'] = 'Nelder-Mead'

    # minimize objective
    result = so.minimize(objective, init, **kwargs)

    # Compute standard deviation if required
    if sd:
        H, nfev_total = hessian(objective, result.x,
                                nfev=result.nfev, f_x0=result.fun)
        Hinv = np.linalg.inv(H)
        x_sd = np.sqrt(np.diagonal(Hinv))
    else:
        nfev_total = result.nfev
        x_sd = np.zeros_like(result.x)

    return result, x_sd, nfev_total


def save_to_backend(backend, content):

    with backend.open("a") as f:
        g = f[backend.name]
        g["keys"] = list(content.keys())
        for key in content:
            g[key] = content[key]

    return


def load_backend(backend):
    """just a shortcut"""
    reader = emcee.backends.HDFBackend(backend, read_only=True)

    storage_dict = {}
    with reader.open() as f:

        g = f[reader.name]
        if 'keys' in g:
            keys = g["keys"][...]

            for key in keys:
                exec('reader.%s = g[key][...]' % key.decode())

    return reader


def remove_backend(backend):
    """just a shortcut"""
    os.remove(backend)
    return


rm_backend = remove_backend
